# Copyright (c) 2024 Roboto Technologies, Inc.
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at https://mozilla.org/MPL/2.0/.

from ._optional import (
    IMPORT_ERROR_MSG_TEMPLATE,
    import_optional_dependency,
)

__all__ = (
    "IMPORT_ERROR_MSG_TEMPLATE",
    "import_optional_dependency",
)
